// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:wyatt_continuous_deployment/src/core/constants/app_constants.dart';
import 'package:wyatt_continuous_deployment/src/core/injection/injectable.dart';
import 'package:wyatt_continuous_deployment/src/core/task/task_cubit.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/context.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/fastlane.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/fastlane/fastlane_deploy_supply_parameters.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/config/get_android_config_usecase.dart';

part 'deploy_android_state.dart';

class DeployAndroidTask extends TaskCubit {
  factory DeployAndroidTask() => DeployAndroidTask._(
        getAndroidConfigUsecase: getIt(),
        fastlane: getIt(),
        checkToolsUsecase: getIt(),
        checkConfigFileUsecase: getIt(),
      );

  DeployAndroidTask._({
    required GetAndroidConfigUsecase getAndroidConfigUsecase,
    required Fastlane fastlane,
    required super.checkToolsUsecase,
    required super.checkConfigFileUsecase,
  })  : _getAndroidConfigUsecase = getAndroidConfigUsecase,
        _fastlane = fastlane;

  final GetAndroidConfigUsecase _getAndroidConfigUsecase;
  final Fastlane _fastlane;

  @override
  Future<void> onRun(Context ctx) async {
    // 1. Get Android Config
    emit(DeployAndroidStateGetAndroidConfig());

    final getAndroidConfigResult =
        await _getAndroidConfigUsecase(ctx.configFile.path);
    if (getAndroidConfigResult.isErr) {
      return emit(
        TaskStateError(
          message: getAndroidConfigResult.err.toString(),
        ),
      );
    }

    // 2. Deploy to Play Store
    emit(DeployAndroidStateDeployToPlayStore());
    final deploySupplyResult = await _fastlane.deployGooglePlay(
      FastlaneDeploySupplyParameters(
        packageName: getAndroidConfigResult.ok!.packageName.toString(),
        track: getAndroidConfigResult.ok!.track,
        serviceAccountPath:
            '${getAndroidConfigResult.ok!.androidFolderPath}/${getAndroidConfigResult.ok!.googleServiceAccountFileName}',
        appBundlePath: AppConstants.androidBundleBuildPath,
        releaseStatus: getAndroidConfigResult.ok!.releaseStatus,
      ),
    );

    if (deploySupplyResult.isErr) {
      return emit(
        TaskStateError(
          message: deploySupplyResult.err.toString(),
        ),
      );
    }
  }
}
