// Copyright (C) 2024 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_continuous_deployment/src/domain/entities/context.dart';
import 'package:wyatt_continuous_deployment/src/domain/usecases/usecases.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

part 'task_state.dart';

abstract class TaskCubit extends Cubit<TaskState> {
  TaskCubit({
    required CheckToolsUsecase checkToolsUsecase,
    required CheckConfigFileUsecase checkConfigFileUsecase,
  })  : _checkToolsUsecase = checkToolsUsecase,
        _checkConfigFileUsecase = checkConfigFileUsecase,
        super(TaskStateInitial());

  final CheckToolsUsecase _checkToolsUsecase;
  final CheckConfigFileUsecase _checkConfigFileUsecase;

  int get status => switch (state) {
        TaskStateFinished(:final exitCode) => exitCode,
        _ => 0,
      };

  /// Whether to check the tools before running the task
  bool get checks => true;

  CheckToolsUsecase get checkToolsUsecase => _checkToolsUsecase;
  CheckConfigFileUsecase get checkConfigFileUsecase => _checkConfigFileUsecase;

  Future<void> awaitForFinished() async {
    if (state is TaskStateFinished) {
      return;
    }

    await for (final state in stream) {
      if (state is TaskStateFinished) {
        return;
      }
    }
  }

  Future<void> run(Context ctx) async {
    if (checks) {
      emit(TaskStateCheckTools());

      final checkToolsResult = await _checkToolsUsecase();
      if (checkToolsResult.isErr) {
        return emit(TaskStateError.fromRes(checkToolsResult));
      }

      emit(TaskStateCheckConfigFile());

      final checkConfigFileResult =
          await _checkConfigFileUsecase(ctx.configFile.path);
      if (checkConfigFileResult.isErr) {
        return emit(TaskStateError.fromRes(checkConfigFileResult));
      }
    }

    // Run the task
    await onRun(ctx);

    if (state is! TaskStateError && state is! TaskStateSucceed) {
      // If no error was emitted, and no success state was emitted,
      // then emit a success state to indicate that the task has finished
      return emit(const TaskStateSucceed());
    }
  }

  Future<void> onRun(Context ctx);
}
